﻿using BenchmarkDotNet.Attributes;
using BenchmarkDotNet.Running;

namespace ListRemoveAllBenchmarks {

	public class RemoveAllTests {
		private const int ItemCount =
#if DONT_RUN_BENCHMARK
			20
#else
			100_000
#endif
			;

		public List<int> sourceList;
		public List<int> iterationList;

		private const int MagicValue = 444;

		[GlobalSetup]
		public void GlobalSetup() {
			sourceList = new List<int>(ItemCount);
			iterationList = new List<int>(ItemCount);

			Random r = new Random(42);
			for (int i = 0; i < ItemCount; i++) {
				int item = r.Next(0, 100);
				if (i == ItemCount / 2) {
					item = MagicValue;
				}

				sourceList.Add(item);
			}
		}

		private void ManualIterationSetup() {
			iterationList.Clear();
			iterationList.AddRange(sourceList);
		}

		[Benchmark]
		public void Remove_Many() {
			ManualIterationSetup();

			var list = iterationList;
			for (int i = 0; i < list.Count; i++) {
				if (list[i] % 2 == 0) {
					list.Remove(list[i]);
					i--;
				}
			}
		}

		[Benchmark]
		public void RemoveAt_Many() {
			ManualIterationSetup();

			var list = iterationList;
			for (int i = 0; i < list.Count; i++) {
				if (list[i] % 2 == 0) {
					list.RemoveAt(i);
					i--;
				}
			}
		}

		[Benchmark]
		public void RemoveAll_Many() {
			ManualIterationSetup();

			var list = iterationList;
			list.RemoveAll(IsEven);
		}

		private static bool IsEven(int x) => x % 2 == 0;

		[Benchmark]
		public void Remove_One() {
			ManualIterationSetup();

			var list = iterationList;
			for (int i = 0; i < list.Count; i++) {
				if (list[i] == MagicValue) {
					list.Remove(list[i]);
					i--;
				}
			}
		}

		[Benchmark]
		public void RemoveAt_One() {
			ManualIterationSetup();

			var list = iterationList;
			for (int i = 0; i < list.Count; i++) {
				if (list[i] == MagicValue) {
					list.RemoveAt(i);
					i--;
				}
			}
		}

		[Benchmark]
		public void RemoveAll_One() {
			ManualIterationSetup();

			var list = iterationList;
			list.RemoveAll(IsMagicValue);
		}

		private static bool IsMagicValue(int x) => x == MagicValue;
	}

	class Program {
		static void Main(string[] args) {
#if !DONT_RUN_BENCHMARK
			BenchmarkRunner.Run<RemoveAllTests>();

#else
			var tests = new RemoveAllTests();
			tests.GlobalSetup();
			Console.WriteLine("Original:");
			tests.sourceList.ForEach(x => Console.WriteLine(x));

			tests.IterationSetup();
			tests.Remove_Many();
			Console.WriteLine("After Remove_Many:");
			tests.iterationList.ForEach(x => Console.WriteLine(x));

			tests.IterationSetup();
			tests.RemoveAt_Many();
			Console.WriteLine("After RemoveAt_Many:");
			tests.iterationList.ForEach(x => Console.WriteLine(x));

			tests.IterationSetup();
			tests.RemoveAll_Many();
			Console.WriteLine("After RemoveAll_Many:");
			tests.iterationList.ForEach(x => Console.WriteLine(x));

			tests.IterationSetup();
			tests.Remove_One();
			Console.WriteLine("After Remove_One:");
			tests.iterationList.ForEach(x => Console.WriteLine(x));

			tests.IterationSetup();
			tests.RemoveAt_One();
			Console.WriteLine("After RemoveAt_One:");
			tests.iterationList.ForEach(x => Console.WriteLine(x));

			tests.IterationSetup();
			tests.RemoveAll_One();
			Console.WriteLine("After RemoveAll_One:");
			tests.iterationList.ForEach(x => Console.WriteLine(x));
#endif
		}
	}
}
